module definitions
    type tracking_cloud
      integer np,ntracks
      real, dimension(:), pointer :: time
      real, dimension(:,:),pointer :: x,y,z,press,temp,velo,veloZ
      integer,dimension(:),pointer :: step
      integer,dimension(:,:),pointer :: tracks
    end type tracking_cloud

end module definitions


!==============================================================================!
!==============================================================================!
!                                                                              |
!              ||===\\                                                         | 
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!==============================================================================!
!==============================================================================!
!                                                                              |
!              TRACKING CLOUD POST PROCESSING   Jan. 2015                      |
!                                                                              |
!==============================================================================!
!==============================================================================!

program post_tracking

use definitions

implicit none

type (tracking_cloud) :: trcl
 
!input parameters from input_tracking_output.txt
character(150) :: input_dir
integer :: output_mode
integer :: output_scaled
integer :: output_first
integer :: output_last
integer :: output_step
integer :: input_first
integer :: input_step
integer :: output_velo
integer :: output_veloz
integer :: output_temp
integer :: output_press
integer :: output_time
integer :: output_istep

!buffer variables for reading
real :: read_time,lref,veloref,rhoref,tempscale
integer :: read_step

!other variables
integer :: i,k,l,err,dummyint
integer :: output_nstep,input_nstep,input_last,output_any
integer :: write_nstep

character(4) :: cistep

write (*,'(a)') '!------------------------------------------------------------------------------|'
write (*,'(a)') '!------------------------------------------------------------------------------|'
write (*,'(a)') '!                                                                              |'
write (*,'(a)') '!          8888888b.   .d88888b.  888     888       d8888 8888888b.            |'
write (*,'(a)') '!          888  "Y88b d88P" "Y88b 888     888      d88888 888   Y88b           |'
write (*,'(a)') '!          888    888 888     888 888     888     d88P888 888    888           |'
write (*,'(a)') '!          888    888 888     888 888     888    d88P 888 888   d88P           |'
write (*,'(a)') '!          888    888 888     888 888     888   d88P  888 8888888P"            |'
write (*,'(a)') '!          888    888 888     888 888     888  d88P   888 888 T88b             |'
write (*,'(a)') '!          888  .d88P Y88b. .d88P Y88b. .d88P d8888888888 888  T88b            |'
write (*,'(a)') '!          8888888P"   "Y88888P"   "Y88888P" d88P     888 888   T88b           |'
write (*,'(a)') '!                                                                              |'
write (*,'(a)') '!             DOUAR-WSMP post-processor for the tracking cloud                 |'
write (*,'(a)') '!------------------------------------------------------------------------------|'
write (*,'(a)') '!------------------------------------------------------------------------------|'
write (*,*) ''

!------------------------------------------------------------------------------|
!    Input parameters
!------------------------------------------------------------------------------|

!Reading input parameters from post-input-file:

open(unit=10,file='input_tracking_outputs.txt',status='old')

read (10,*) input_dir
read (10,*) output_mode
read (10,*) output_scaled
read (10,*) output_first
read (10,*) output_last
read (10,*) output_step
read (10,*) input_first
read (10,*) input_step
read (10,*)  !here come the age parameters
read (10,*)  !here come the track parameters
read (10,*) output_velo
read (10,*) output_veloz
read (10,*) output_temp
read (10,*) output_press
read (10,*) output_time
read (10,*) output_istep

close (10)

!calculating and andjusting input and output timesteps
output_nstep = (output_last-output_first)/output_step+1
output_last = output_first+(output_nstep-1)*output_step

input_last = output_last
input_nstep = (input_last-input_first)/input_step+1

!writign output parameters to screen
write (*,*) 'Input Directory: ',input_dir
write (*,'(a,i2,a,i3,a,i3,a,i3)') 'Input Timesteps: Reading ',input_nstep,' input files & 
from step ',input_first,' to ',input_last,' in steps of ',input_step
write (*,'(a,i2,a,i3,a,i3,a,i3)') 'Output Timesteps: Generating ',output_nstep,' output files & 
from step ',output_first,' to ',output_last,' in steps of ',output_step
write (*,'(a,i1)') 'Output mode: ',output_mode
write (*,'(a,l)') 'Write natural units:     ',(output_scaled==1)
write (*,'(a,l)') 'Write track velocity:    ',(output_velo==1)
write (*,'(a,l)') 'Write track z velocity:  ',(output_veloz==1)
write (*,'(a,l)') 'Write track temperature: ',(output_temp==1)
write (*,'(a,l)') 'Write track pressure:    ',(output_press==1)
write (*,'(a,l)') 'Write track time:        ',(output_time==1)
write (*,'(a,l)') 'Write track step number: ',(output_istep==1)
write (*,'(a)') ''
write (*,'(a)') '!------------------------------------------------------------------------------|'
write (*,'(a)') ''
!------------------------------------------------------------------------------|
!    Reading Tracking Cloud Files
!------------------------------------------------------------------------------|

!reading input file header for array sizes
call int_to_char (cistep,4,input_last)
write (*,*) 'last input file: '//trim(input_dir)//'tracking_t'//cistep//'.bin'

open (unit=20,file=trim(input_dir)//'tracking_t'//cistep//'.bin',status='old',form='unformatted')
read (20) trcl%np,trcl%ntracks,read_step,read_time
read (20) lref,veloref,rhoref,tempscale
close (20)

!allocating master arrays for input
allocate (trcl%x(trcl%np,input_nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%x$')
allocate (trcl%y(trcl%np,input_nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%y$')
allocate (trcl%z(trcl%np,input_nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%z$')
allocate (trcl%temp(trcl%np,input_nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%temp$')
allocate (trcl%press(trcl%np,input_nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%press$')
allocate (trcl%step(input_nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%step$')
allocate (trcl%time(input_nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%time$')
allocate (trcl%tracks(trcl%ntracks,3),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%tracks$')

!read all input files and save in collective arrays
k = 1
do i = input_first,input_last,input_step
    !read all information for this time step
    call int_to_char (cistep,4,i)
    open (unit=20,file=trim(input_dir)//'tracking_t'//cistep//'.bin',status='old',form='unformatted')
    
    read (20) dummyint,dummyint,trcl%step(k),trcl%time(k)
    read (20) 
    read (20) trcl%x(:,k)
    read (20) trcl%y(:,k)
    read (20) trcl%z(:,k)
    read (20) trcl%temp(:,k)
    read (20) trcl%press(:,k)
    
    if (i==input_last) then
        do l=1,trcl%ntracks
            read (20) trcl%tracks(l,:)
        end do
        !write (*,*) k, input_nstep
    end if

    !advance column in master array
    k = k+1

    close (20)
end do

write (*,'(a)') 'Finished reading input files'

!adjust track timesteps to actual position in master array
!WARNING: this aggregates tracks from the timesteps in between
!how to deal with them properly?
do k=2,3
   do i=1,trcl%ntracks
        trcl%tracks(i,k) = max(1,(trcl%tracks(i,k)-input_first)/input_step + 1)
   end do
end do


!------------------------------------------------------------------------------|
!    Scaling Values to natural units
!------------------------------------------------------------------------------|
!check what age prediction needs...
!units of scaling parameters:
!lref: km
!veloref: mm/yr = km/Myr
!rhoref: kg/m^3
!tempscale: °C

if (output_scaled==1) then
    do k=1,input_nstep
        do i=1,trcl%np
            trcl%x(i,k) = trcl%x(i,k)*lref !pos in km
            trcl%y(i,k) = trcl%y(i,k)*lref
            trcl%z(i,k) = trcl%z(i,k)*lref
            trcl%temp(i,k) = trcl%temp(i,k)*tempscale !temp in C
            trcl%press(i,k) = trcl%press(i,k)*1.0e-3*9.81*rhoref*lref !pressure in MPa
        end do
        trcl%time(k) = trcl%time(k)*lref/veloref !time in Myr
    end do
end if


!------------------------------------------------------------------------------|
!    Calculating Output values
!------------------------------------------------------------------------------|

!track values
velo: if (output_velo==1) then
    allocate (trcl%velo(trcl%np,input_nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%velo$')
    do i=1,trcl%np
        trcl%velo(i,1) = abs( & 
            ((trcl%x(i,2)-trcl%x(i,1))**2+(trcl%y(i,2)-trcl%y(i,1))**2+(trcl%z(i,2)-trcl%z(i,1))**2) &
            /(trcl%time(2)-trcl%time(1)))
    end do
    do i=1,trcl%np
        do k=2,input_nstep
            trcl%velo(i,k) = abs( & 
                ((trcl%x(i,k)-trcl%x(i,k-1))**2+(trcl%y(i,k)-trcl%y(i,k-1))**2+(trcl%z(i,k)-trcl%z(i,k-1))**2) &
                /(trcl%time(k)-trcl%time(k-1)))
        end do
    end do
end if velo


veloz: if (output_veloz==1) then
    allocate (trcl%veloz(trcl%np,input_nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%veloz$')
    do i=1,trcl%np
        trcl%velo(i,1) = ((trcl%z(i,2)-trcl%z(i,1))**2)/(trcl%time(2)-trcl%time(1))
    end do
    do i=1,trcl%np
        do k=2,input_nstep
            trcl%velo(i,k) = ((trcl%z(i,k)-trcl%z(i,k-1))**2)/(trcl%time(k)-trcl%time(k-1))
        end do
    end do
end if veloz

!------------------------------------------------------------------------------|
!    Checking for NaNs and variable ranges
!------------------------------------------------------------------------------|


do k=1,input_nstep
    if (isnan(trcl%time(k))) trcl%time(k)=1.0e38
    if (1.0e37<trcl%time(k)) trcl%time(k)=1.0e37
    if (-1.0e37<trcl%time(k)) trcl%time(k)=-1.0e37
    if (1.0e-37>abs(trcl%time(k))) trcl%time(k)=0.0
    do i=1,trcl%np
        if (isnan(trcl%temp(i,k))) trcl%temp(i,k)=1.0e38
        if (1.0e37<trcl%temp(i,k)) trcl%temp(i,k)=1.0e37
        if (-1.0e37<trcl%temp(i,k)) trcl%temp(i,k)=-1.0e37
        if (1.0e-37>abs(trcl%temp(i,k))) trcl%temp(i,k)=0.0
        
        if (isnan(trcl%press(i,k))) trcl%press(i,k)=1.0e38
        if (1.0e37<trcl%press(i,k)) trcl%press(i,k)=1.0e37
        if (-1.0e37<trcl%press(i,k)) trcl%press(i,k)=-1.0e37
        if (1.0e-37>abs(trcl%press(i,k))) trcl%press(i,k)=0.0

        if (output_velo==1) then
            if (isnan(trcl%velo(i,k))) trcl%velo(i,k)=1.0e38
            if (1.0e37<trcl%velo(i,k)) trcl%velo(i,k)=1.0e37
            if (-1.0e37<trcl%velo(i,k)) trcl%velo(i,k)=-1.0e37
            if (1.0e-37>abs(trcl%velo(i,k))) trcl%velo(i,k)=0.0
        end if

        if (output_veloz==1) then
            if (isnan(trcl%veloz(i,k))) trcl%veloz(i,k)=1.0e38
            if (1.0e37<trcl%veloz(i,k)) trcl%veloz(i,k)=1.0e37
            if (-1.0e37<trcl%veloz(i,k)) trcl%veloz(i,k)=-1.0e37
            if (1.0e-37>abs(trcl%veloz(i,k))) trcl%veloz(i,k)=0.0
        end if
    end do
end do

!------------------------------------------------------------------------------|
!    Writing Output Files
!------------------------------------------------------------------------------|
output_any = output_velo+output_veloz+output_temp+output_press+output_time+output_istep

if (output_mode==3) then
!output all tracks
    do l=output_first,output_last,output_step
        call int_to_char (cistep,4,l)
        open (unit=30,file=trim(input_dir)//'alltracks_t'//cistep//'.vtk',status='replace')

        !calculate part of array to write
        write_nstep = (l-input_first)/input_step+1

        !write header
        write(30,'(a)') '# vtk DataFile Version 3.1'
        write(30, '(a)') 'DOUAR tracking cloud output of all tracks'
        write(30, '(a)') 'ASCII'
        write(30, '(a)') 'DATASET UNSTRUCTURED_GRID'
        write(30, '(a)') ''

        !write point coordinates
        write(30, '(a,i,a)') 'POINTS ', trcl%np*write_nstep, ' FLOAT'
        do i=1,trcl%np
            do k=1,write_nstep
                write(30,'(3ES12.5E2)') trcl%x(i,k),trcl%y(i,k),trcl%z(i,k)
            end do
        end do
        write(30, '(a)') ''

        !write cell data
        write(30, '(a,i10,i10)') 'CELLS ',trcl%np, trcl%np*(write_nstep+1)
        do i = 0,trcl%np-1
            write(30, '(i)',advance='no') write_nstep
            do k = 1,write_nstep
                write(30, '(i)', advance='no') i*write_nstep+k-1
            end do
            write (30, '(a)') '' 
        end do
        write (30, '(a)') '' 

        !write cell data type, 4=polyline
        write(30, '(a,i10)') 'CELL_TYPES ',trcl%np
        do i = 1,trcl%np
            write(30, '(i2)') 4
        end do
        write (30, '(a)') '' 

        !write point datasets
        if (output_any > 0) write (30,'(a,i)') 'POINT_DATA ',trcl%np*write_nstep
        
        if (output_velo==1) then
            write (30,'(a)') 'SCALARS Velo FLOAT 1'
            write (30,'(a)') 'LOOKUP_TABLE default'
            do i=1,trcl%np
                do k=1,write_nstep
                    write(30,'(ES)') trcl%velo(i,k)
                end do
            end do
            write(30, '(a)') ''
        end if

        if (output_veloz==1) then
            write (30,'(a)') 'SCALARS VeloZ FLOAT 1'
            write (30,'(a)') 'LOOKUP_TABLE default'
            do i=1,trcl%np
                do k=1,write_nstep
                    write(30,'(ES12.5E2)') trcl%veloz(i,k)
                end do
            end do
            write(30, '(a)') ''
        end if

        if (output_temp==1) then
            write (30,'(a)') 'SCALARS Temperature FLOAT 1'
            write (30,'(a)') 'LOOKUP_TABLE default'
            do i=1,trcl%np
                do k=1,write_nstep
                    write(30,'(ES12.5E2)') trcl%temp(i,k)
                end do
            end do
            write(30, '(a)') ''
        end if

        if (output_press==1) then
            write (30,'(a)') 'SCALARS Pressure FLOAT 1'
            write (30,'(a)') 'LOOKUP_TABLE default'
            do i=1,trcl%np
                do k=1,write_nstep
                    write(30,'(ES12.5E2)') trcl%press(i,k)
                end do
            end do
            write(30, '(a)') ''
        end if

        if (output_time==1) then
            write (30,'(a)') 'SCALARS Time FLOAT 1'
            write (30,'(a)') 'LOOKUP_TABLE default'
            do i=1,trcl%np
                do k=1,write_nstep
                    write(30,'(ES12.5E2)') trcl%time(k)
                end do
            end do
            write(30, '(a)') ''
        end if

        if (output_istep==1) then
            write (30,'(a)') 'SCALARS Step INTEGER 1'
            write (30,'(a)') 'LOOKUP_TABLE default'
            do i=1,trcl%np
                do k=1,write_nstep
                    write(30,'(i)') trcl%step(k)
                end do
            end do
            write(30, '(a)') ''
        end if

        close (30)
    end do
end if





deallocate (trcl%x)   
deallocate (trcl%y)
deallocate (trcl%z)
deallocate (trcl%temp)
deallocate (trcl%press)
deallocate (trcl%step)
deallocate (trcl%time)
deallocate (trcl%tracks)
if (output_velo==1) deallocate (trcl%velo)
if (output_veloz==1) deallocate (trcl%veloz)


end
